// -*- Mode: vala; indent-tabs-mode: nil; tab-width: 4 -*-
/*-
 * Copyright (c) 2013-2016 elementary LLC.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * Authored by: Corentin Noël <corentin@elementary.io>
 */

// This is compatible with Ubuntu Online Accounts.
public class OnlineAccounts.Keyring : GSignond.SecretStorage {
    public enum SignonSecretType {
        NOTYPE = 0,
        PASSWORD,
        USERNAME,
        DATA
    }

    Secret.Schema schema = null;
    Error error = null;
    public override bool open_db () {
        if (schema == null) {
            schema = new Secret.Schema ("org pantheon.OnlineAccounts.Secrets", Secret.SchemaFlags.DONT_MATCH_NAME,
                                     "signon-type", Secret.SchemaAttributeType.INTEGER,
                                     "signon-id", Secret.SchemaAttributeType.INTEGER,
                                     "signon-method", Secret.SchemaAttributeType.INTEGER);
        }

        return true;
    }

    public override bool close_db () {
        return true;
    }

    public override bool clear_db () {
        return true;
    }

    public override bool is_open_db () {
        return (schema != null);
    }

    public override GSignond.Credentials load_credentials (uint32 id) {
        var credidential = new GSignond.Credentials ();
        string username;
        string password;
        load_secret (SignonSecretType.USERNAME, id, 0, out username);
        load_secret (SignonSecretType.PASSWORD, id, 0, out password);
        credidential.set_data (id, username, password);
        return credidential;
    }

    public override bool update_credentials (GSignond.Credentials creds) {
        var id = creds.get_id ();
        var password = creds.get_password ();
        if (password != null && password != "") {
            if (!store_secret (SignonSecretType.PASSWORD, id, 0, password)) {
                return false;
            }
        }

        var username = creds.get_username ();
        if (username != null && username != "") {
            if (!store_secret (SignonSecretType.USERNAME, id, 0, username)) {
                return false;
            }
        }

        return true;
    }

    public override bool remove_credentials (uint32 id) {
        try {
            return Secret.password_clear_sync (schema, null, "signon-id", id);
        } catch (Error e) {
            critical (e.message);
            error = e;
            return false;
        }
    }

    public override bool check_credentials (GSignond.Credentials creds) {
        return base.check_credentials (creds);
    }

    public override GSignond.Dictionary load_data (uint32 id, uint32 method) {
        var result = new GSignond.Dictionary ();
        string data_serialized;
        load_secret (SignonSecretType.DATA, id, method, out data_serialized);
        if (data_serialized == null) {
            return result;
        }

        try {
            var variant = Variant.parse (null, data_serialized);
            result = new GSignond.Dictionary.from_variant (variant);
        } catch (Error e) {
            critical (e.message);
            error = e;
        }

        return result;
    }

    public override bool update_data (uint32 id, uint32 method, GSignond.Dictionary data) {
        string secret = "%s".printf (data.to_variant ().print (true));
        return store_secret (SignonSecretType.DATA, id, method, secret);
    }

    public override bool remove_data (uint32 id, uint32 method) {
        try {
            return Secret.password_clear_sync (schema, null, "signon-id", id, "signon-method", method, null);
        } catch (Error e) {
            critical (e.message);
            error = e;
            return false;
        }
    }

    public override GLib.Error get_last_error () {
        return error;
    }

    public bool store_password (int type, int id, int method, string password) {
        if (password == null || password == "")
            return false;

        try {
            return Secret.password_store_sync (schema, Secret.COLLECTION_DEFAULT, "Online Account",
                                        password, null, "signon-type", type, "signon-id", id,
                                        "signon-method", method, null);
        } catch (Error e) {
            critical (e.message);
            error = e;
            return false;
        }
    }

    public bool store_secret (SignonSecretType type, uint32 id, uint32 method, string secret) {
        if (secret == null || secret == "")
            return false;
        var display_name = "Online Account: id %u-%u".printf (id, type);
        string? signonMethod = (type == SignonSecretType.DATA) ? "signon-method" : null;
        try {
            Secret.password_store_sync (schema, Secret.COLLECTION_DEFAULT, display_name, secret, null,
                                        "signon-type", type,
                                        "signon-id", id,
                                        signonMethod, method);
        } catch (Error e) {
            critical (e.message);
            error = e;
            return false;
        }

        return true;
    }

    public bool load_secret (SignonSecretType type, uint32 id, uint32 method, out string secret) {
        string? signonMethod = (type == SignonSecretType.DATA) ? "signon-method" : null;
        try {
            string data = Secret.password_lookup_sync (schema, null, "signon-type", type,
                                                                     "signon-id", id,
                                                                     signonMethod, method);
            secret = data;
        } catch (Error e) {
            warning (e.message);
            secret = null;
            error = e;
            return false;
        }

        return true;
    }
}
